/* serial.c								*/
/*									*/
/* Copyright (C) 1994 by Robert Armstrong				*/
/*									*/
/* This program is free software; you can redistribute it and/or modify	*/
/* it under the terms of the GNU General Public License as published by	*/
/* the Free Software Foundation; either version 2 of the License, or	*/
/* (at your option) any later version.					*/
/*									*/
/* This program is distributed in the hope that it will be useful, but	*/
/* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-	*/
/* ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General	*/
/* Public License for more details.					*/
/*									*/
/* You should have received a copy of the GNU General Public License	*/
/* along with this program; if not, visit the website of the Free	*/
/* Software Foundation, Inc., www.gnu.org.				*/
/*									*/
/*									*/
/*   This file contains an interface to the standard PC COM port	*/
/* hardware.  It's intended for use with the TU58 device driver, but	*/
/* it is semi-general purpose.  The receiver is interrupt driven and	*/
/* a circular buffer is used to hold bytes until they can be processed.	*/
/* The transmitter is also interrupt driven - this allows the code to	*/
/* compute the next buffer of data while the current one is sent. There	*/
/* is no special support for modem control (the TU58 doesn't use it!).	*/
/*									*/
/* These functions are provided:					*/
/*									*/
/*      COMPut	  - add one byte to the transmitter buffer		*/
/*      COMPutW	  - like Put(), but wait if the buffer is full		*/
/*      COMGet	  - get one byte from the receive buffer		*/
/*      COMGetW   - like Get(), but wait with a timeout			*/
/*      COMFormat - set data format (parity, number of bits, etc)	*/
/*      COMBaud   - set the baud rate					*/
/*      COMBreak  - force a break condition on the TxD signal		*/
/*      COMOpen   - open the port and seize the interrupt vector	*/
/*      COMClose  - close the port and restore the interrupt		*/
/*									*/
/*									*/
/*   There are several chips that can be used in PC serial ports - the	*/
/* 8250 and 8250A, the 16450, the 16550 and 16550A.  The 8250 is the	*/
/* "traditional" chip used in the original PC. It is now rarely used in	*/
/* new hardware, but remains the lowest common denominator. The 8250 is	*/
/* expecially annoying to programmers because it contains a number of	*/
/* hardware bugs which we must program around!  The 16450 is simply a	*/
/* faster version of the 8250, and as far as the programmer knows they	*/
/* are the same (although the 16450 fixes most of the 8250's bugs!).	*/
/* The 16550 is a 16450 with a FIFO buffer added. Unfortunately the	*/
/* 16550 FIFO doesn't really work - it will generate spurious bytes and	*/
/* as such isn't safe to use!						*/
/*									*/
/*  The 16550A fixes the FIFO problems and actually works great. Legend	*/
/* has it that National Semiconductor was so embarassed about the bugs	*/
/* in the 16550 that they added	a special bit to the 16550A FCR to make	*/
/* it easy to recognize.  Whatever the reason, this bit does exist and	*/
/* we make use of it to enable the FIFO on 16550A UARTs.		*/
/*									*/
/*					Bob Armstrong [15-Aug-94]	*/


#include <stdio.h>		/* we need this for NULL		*/
#include <dos.h>		/* used only for MK_FP !!		*/
#include "standard.h"		/* all definitions for this project!	*/
#include "serial.h"		/* PC serial port hardware definitions	*/


/* BIOS data segement areas related to the COM ports...			*/
PRIVATE WORD far *Ports		/* BIOS table of COM port I/O addresses	*/
		= MK_FP(0x40, 0);
PRIVATE LONG far volatile *Ticks/* BIOS location which counts at 18.2Hz	*/
		= MK_FP(0x40, 0x6C);

/* These variables describe the port used and its settings...		*/
PRIVATE WORD      port = 0;	/* I/O base address for this COM port	*/
PRIVATE WORD	  vector;	/* interrupt vector used by the port	*/
PRIVATE BYTE	  picmask;	/* PIC mask for this COM port interrupt	*/
PRIVATE void far *old_isr;	/* previous interrupt service routine	*/


/* Items exported by the interrupt service module, UARTISR.ASM...	*/
extern BYTE volatile txbuf[];	/* circular buffer for transmitting	*/
extern BYTE volatile rxbuf[];	/*  "   "      "    "  receiving	*/
extern WORD volatile txputp;	/* next byte to fill in xmit buffer	*/
extern WORD volatile txgetp;	/*  "    "   "  empty "  "     "	*/
extern WORD volatile rxputp;	/* next byte to fill in receive buffer	*/
extern WORD volatile rxgetp;	/*  "    "   "  empty "   "       "	*/
extern WORD	     rxbufsiz;	/* size of the receiver buffer		*/
extern WORD	     txbufsiz;	/*  "   "   "  transmitter "		*/
extern void near recv_isr (void);	/* receiver interrupt service	*/
extern void near xmit_isr (void);	/* transmitter  "        "	*/
extern void near uart_isr (void);	/* UART	        "	 "	*/
extern void near setup_isr(WORD, WORD);	/* installs the ISR routines	*/


/*   Receiver interrupts are always enabled, but the transmitter	*/
/* interrupt enable must be turned off whenever we have nothing left in	*/
/* the buffer to send.  When something is next added to the buffer we	*/
/* have to be sure to re-enable the transmitter to get things started.	*/
/* But at the same time we have to be careful not to mess things up if	*/
/* the transmitter is currently active.  This routine takes care of all	*/
/* those little problems..						*/
PRIVATE void start_transmitter (void)
{
  BYTE lsr;
  INT_OFF;  IOSETBIT(port+IER, IER_TBE);
  /* Some UART chips are a little slow - give them a chance to catch up. */
  asm nop; asm nop; asm nop; asm nop;
  INBYTE(port+LSR, lsr);
  if ((lsr & LSR_TXE) != 0)  xmit_isr();
  INT_ON;
}


/* COMGet */
/*   This routine will return the next byte from the receiver circular	*/
/* buffer.  If the buffer is empty, it will return a function value of	*/
/* FALSE.  It will never block (wait), no matter what.			*/
/*									*/
/* OUTPUT:								*/
/*	ch	- the 8 bit byte from the queue				*/
/*									*/
/* RETURNS:	- TRUE if a byte was returned, FALSE if there were none	*/
/*									*/
PUBLIC BOOLEAN COMGet (BYTE far *ch)
{
  INT_OFF;
    if (rxgetp == rxputp)  {INT_ON;  return FALSE;}
    if (++rxgetp >= rxbufsiz)  rxgetp = 0;
    *ch = rxbuf[rxgetp];
  INT_ON;
  return TRUE;
}


/* COMGetW */
/*   This routine is identical to COMGet, except that it will wait in	*/
/* the even that the buffer is empty.  The wait can have a timeout, or	*/
/* it can simply wait forever.  The timeout is specified in PC timer	*/
/* ticks (approximately 55mS) and may be up to 32,767 (a little over	*/
/* 30 minutes!).  If a character is found then TRUE is returned - if	*/
/* the timeout expires with no character then FALSE results.  A timeout	*/
/* of 0 (or less) will wait forever.					*/
/*									*/
/* INPUTS:								*/
/*	timeout	- timeout value, in ticks.  Use 0 to wait forever!	*/
/*									*/
/* OUTPUTS:								*/
/*	ch	- the byte extracted from the receiver queue		*/
/*									*/
/* RETURNS:	- TRUE if a byte was read - FALSE if it timed out	*/
/*									*/
PUBLIC BOOLEAN COMGetW (BYTE far *ch, WORD timeout)
{
  LONG time = *Ticks;
  while (!COMGet(ch)) {
    if (timeout > 0  &&  time != *Ticks) {
      if (--timeout <= 0)  return FALSE;
      time = *Ticks;
    }
  }
  return TRUE;
}


/* COMPeekW */
/*   This routine also returns the next character from the receiver	*/
/* buffer, but the buffer pointer is not advanced.  This means that the	*/
/* next call to COMGet will return the same character.  It provides a	*/
/* simple one character look ahead function.  It also provides a timed	*/
/* wait function, like COMGetW.						*/
/*									*/
/* INPUTS:								*/
/*	timeout	- timeout value, in ticks.  Use 0 to never wait!!	*/
/*									*/
/* OUTPUTS:								*/
/*	ch	- the byte extracted from the receiver queue		*/
/*									*/
/* RETURNS:	- TRUE if a byte was read - FALSE if it timed out	*/
/*									*/
PUBLIC BOOLEAN COMPeekW (BYTE far *ch, WORD timeout)
{
  LONG time = *Ticks;  WORD getp;
  while (TRUE) {
    INT_OFF;
    if (rxgetp != rxputp)  {
      if ((getp=rxgetp+1) >= rxbufsiz)  getp = 0;
      *ch = rxbuf[getp];
      INT_ON;
      return TRUE;
    }
    INT_ON;
    if (timeout == 0)  return FALSE;
    if (time != *Ticks)
      {  --timeout;  time = *Ticks;  }
  }
}


/* COMPut */
/*   This routine will add a byte to the transmitter circular buffer.	*/
/* If the buffer is full, then the byte is not added and it will return	*/
/* FALSE.  It never waits for space to become available!  If we haven't	*/
/* transmitted anything for a while, then the transmitter interrupts	*/
/* will have been turned off.  In that case it's important that we also	*/
/* fire up the transmitter again!					*/
/*									*/
/* INPUTS:								*/
/*	ch	- the byte to be transmitted (eventually!)		*/
/*									*/
/* RETURNS:	TRUE if the byte was queued, FALSE if there was no room	*/
/*									*/
PUBLIC BOOLEAN COMPut (BYTE ch)
{
  int put;
  INT_OFF;
    if ((put = txputp+1)  >=  txbufsiz)  put = 0;
    if (put == txgetp)  {INT_ON;  return FALSE;}
    txbuf[ txputp = put ] = ch;
  INT_ON;
  start_transmitter();
  return TRUE;
}


/* COMPutW */
/*   This routine is identical to COMPut, except that it will wait	*/
/* for buffer space if necessary.  There is no timeout in this case,	*/
/* since we can be relatively confident (barring software bugs!) that	*/
/* the UART will eventually catch up.					*/
/*									*/
/* INPUTS:								*/
/*	ch	- the byte to be transmitted				*/
/*									*/
PUBLIC void COMPutW (BYTE ch)
{
  while (!COMPut(ch))  continue;
}


/* COMBreak */
/*   This routine will transmit a RS-232 break.  This isn't a character	*/
/* at all, but requires forcing the line into a spacing state for a 	*/
/* long (at least 100mS) period.  This routine generates the delay by	*/
/* watching the BIOS timer tick interrupt (approximately 55mS).		*/
/*									*/
/*   IMPORTANT!  This routine waits for the transmitter queue to empty	*/
/* before starting.  This routine will flush the receiver queue before	*/
/* returning!  The former ensures that the break is synchronized with	*/
/* the data.  The latter ensures that it doesn't add any junk.		*/
/*									*/
/* INPUTS:	(none)							*/
/*									*/
PUBLIC void COMBreak (void)
{
  int i;  LONG time;
  while (txgetp != txputp)  continue;
  IOSETBIT(port+LCR, LCR_BREAK);
  for (i = 0;  i < 3;  ++i) {
    time = *Ticks;
    while (time == *Ticks)  continue;
  }
  INT_OFF;
    rxgetp = rxputp = 0;
    IOCLRBIT(port+LCR, LCR_BREAK);
  INT_ON;
}


/* COMBaud */
/*   This routine will set the baud rate for the serial port.  All the	*/
/* UARTs have an internal baud rate generator (simply a programmable	*/
/* divisor) which counts at 115,200Hz.  Setting the baud rate is simply	*/
/* a matter of computing the appropriate divisor.  Any baud rate from	*/
/* 1 to 65536 BPS can be set, even non-standard ones!  The limit of	*/
/* 65,535 BPS is because we use word sized arithmetic here.  An illegal	*/
/* baud rate will be ignored without changing the current UART speed.	*/
/*									*/
/* INPUTS:								*/
/*	baud	- desired baud rate, from 1 to 64K BPS			*/
/*									*/
PUBLIC void COMBaud (WORD baud)
{
  BYTE dll, dlh;  WORD divisor;  LONG maxbaud = MAXBAUD;
  if (baud == 0)  return;
  /*   This assembly divides a 32 bit long (MAXBAUD) by a 16 bit word	*/
  /* (baud) and produces a 16 bit result.  It's done this way to avoid	*/
  /* having the compiler generate a call to the C RTL LDIV function.	*/
  asm MOV BX, baud;
  asm MOV DX, WORD PTR maxbaud+2;
  asm MOV AX, WORD PTR maxbaud;
  asm DIV BX;
  asm MOV dll, AL;
  asm MOV dlh, AH;
  /*   Now load the divisor latches in the UART.  It would be very bad	*/
  /* if an interrupt happened while the DLAB bit is set!		*/
  INT_OFF;  IOSETBIT(port+LCR, LCR_DLAB);
  OUTBYTE(port+DLL, dll);  OUTBYTE(port+DLH, dlh);
  IOCLRBIT(port+LCR, LCR_DLAB);  INT_ON;
}


/* COMFormat */
/*   This routine sets the character format used by the UART, including	*/
/* the number of data bits, the parity, and the number of stop bits.	*/
/* The character format is specified in the "natural" form as a three	*/
/* character ASCII string.  Note that any illegal format string is	*/
/* ignored without changing the current settings.			*/
/*									*/
/* INPUTS:								*/
/*	format	- 3 character ASCIZ string of the form "wps", where:	*/
/*			"w" is the word length - 5, 6, 7, or 8		*/
/*			"p" is the parity - n (none), e (even), o (odd)	*/
/*					    m (mark), or s (space)	*/
/*			"s" is the stop bits - 1 or 2			*/
/*									*/
/*		   e.g., "8n1" is 8 data bits, no parity, 1 stop bit	*/
/*		 	 "7m2" is 7 data bits, mark parity, 2 stop bits	*/
/*									*/
PUBLIC void COMFormat (char *format)
{
  BYTE lcr = 0;
  if (format == NULL)  return;
  /* Decode the word length - 5, 6, 7, or 8... */
  switch (*format++) {
    case '5':  lcr |= LCR_5BITS;  break;
    case '6':  lcr |= LCR_6BITS;  break;
    case '7':  lcr |= LCR_7BITS;  break;
    case '8':  lcr |= LCR_8BITS;  break;
    default:   return;
  }
  /* Decode the parity - none, even, odd, mark or space... */
  switch (*format++) {
    case 'o':  case 'O':  lcr |= LCR_PARODD;  break;
    case 'e':  case 'E':  lcr |= LCR_PAREVE;  break;
    case 'm':  case 'M':  lcr |= LCR_PARMAR;  break;
    case 's':  case 'S':  lcr |= LCR_PARSPA;  break;
    default:  return;
  }
  /* Lastly decode the number of stop bits - 1 or 2... */
  switch (*format++) {
    case '1':                     break;
    case '2':  lcr |= LCR_2STOP;  break;
    default:  return;
  }
  /* That has to be the end of the string.
  if (*format != '\0')  return;
  /* Set the line characteristics register in the UART and we're done. */
  OUTBYTE(port+LCR, lcr);
}


/* COMClose */
/*   This routine will close the serial port if it is opened.  It will	*/
/* disable interrupts, restore the original interrupt vector, and reset	*/
/* the modem control outputs.  Note that it also always disables the	*/
/* FIFO on UARTs so equipped - leaving the FIFO on can cause some older	*/
/* software to fail!							*/
/*									*/
/* INPUTS:	(none)							*/
/*									*/
PUBLIC void COMClose (void)
{
  if (port == 0)  return;
  OUTBYTE(port+IER, 0);  OUTBYTE(port+MCR, 0);  OUTBYTE(port+FCR, 0);
  IOSETBIT(PIC_IMR, picmask);
  SETVECTOR(vector, old_isr);
  port = 0;
}


/* COMOpen */
/*   This routine attaches a serial port and prepares it for our I/O.	*/
/* It is passed the port number, 1..4.  If the port number is out of	*/
/* this range, or if the specified port does not exist on this PC, then	*/
/* FALSE is returned.  Otherwise it initializes the UART, grabs the	*/
/* interrupt vector, clears the buffers, and enables interrupts.	*/
/*									*/
/* INPUTS:								*/
/*	com_port - port number to be used [1..4]			*/
/*									*/
/* RETURNS:	FALSE if the port is invalid - TRUE otherwise		*/
/*									*/
PUBLIC BOOLEAN COMOpen (int com_port)
{
  BYTE iir;

  /* If a port is already open, close it first! */
  if (com_port < 1  ||  com_port > 4)  return FALSE;
  if (port != 0)  COMClose();

  /* Setup the port parameters - address, vector and PIC mask... */
  if ((port = Ports[com_port-1]) == 0) return FALSE;
  if ((com_port & 1) != 0)
    vector = COM1VEC,  picmask = IMR_COM1;
  else
    vector = COM2VEC,  picmask = IMR_COM2;

  /* Initialize all the circular buffers to be empty... */
  txputp = txgetp = rxputp = rxgetp = 0;

  /*   Always assert the DTR and RTS modem signals (we don't use them,	*/
  /* but just in case anybody out there cares)...			*/
  OUTBYTE(port+MCR, MCR_DTR | MCR_RTS);

  /*   The 16550 UARTs have a hardware FIFO that can dramatically cut	*/
  /* our overhead at high baud rates.  Unfortunately the FIFO in the	*/
  /* original 16550 chip had a hardware bug and would generate spurious	*/
  /* characters!  The 16550A version of the chip fixed this problem.	*/
  /* National Semiconductor was so embarrassed about the original bug	*/
  /* that they made it possible to detect the working version of the	*/
  /* chip by testing for BOTH the FEN1 and FEN2 bits set in the IIR.	*/
  OUTBYTE(port+FCR, FCR_ENB | FCR_F14 | FCR_RFR | FCR_TFR);
  INBYTE(port+IIR, iir);
  if ((iir & IIR_FEN1) == 0  ||  (iir & IIR_FEN2) == 0) {
    OUTBYTE(port+FCR, 0);  /* This chip is not a 16550A !*/
  } else
    cprintf("FIFO enabled.\n");

  /* Save the old interrupt vector, and point to our ISR... */
  GETVECTOR(vector, old_isr);
  setup_isr(vector, port);

  /*   Now enable UART interrupts.  Note that the PC hardware designers	*/
  /* used the GP02 output as a kind of "local" interrupt enable, in the	*/
  /* case of more than one serial port on the same interrupt level.	*/
  /* At the moment, we only enable receiver interrupts.  Transmitter	*/
  /* interrupts will be enabled when something is put in the buffer.	*/
  IOSETBIT(port+MCR, MCR_GPO2);  OUTBYTE(port+IER, IER_RBF);
  IOCLRBIT(PIC_IMR, picmask);

  return TRUE;
}
