/* tu58.c								*/
/*									*/
/* Copyright (C) 1994 by Robert Armstrong				*/
/*									*/
/* This program is free software; you can redistribute it and/or modify	*/
/* it under the terms of the GNU General Public License as published by	*/
/* the Free Software Foundation; either version 2 of the License, or	*/
/* (at your option) any later version.					*/
/*									*/
/* This program is distributed in the hope that it will be useful, but	*/
/* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-	*/
/* ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General	*/
/* Public License for more details.					*/
/*									*/
/* You should have received a copy of the GNU General Public License	*/
/* along with this program; if not, visit the website of the Free	*/
/* Software Foundation, Inc., www.gnu.org.				*/
/*									*/
/*									*/
/*   This module contains all the code necessary to handle the serial	*/
/* communications protocol with the TU58 controller. It provides a high	*/
/* level, function oriented interface to the DOS device driver.  The	*/
/* TU58 protocol is actually relatively simple, mostly because the TU58	*/
/* only implements a handful of functions!				*/
/*									*/
/*   The functions provided are:					*/
/*									*/
/*	TUInitialize - establish two way communications with the drive	*/
/*	TUDiagnose   - ask the drive to execute its built in self test	*/
/*	TUSend       - send raw packets to the drive			*/
/*	TUReceive    - receive raw packets from the drive		*/
/*	TUEnd        - wait for and receive an acknowledgement packet	*/
/*	TUPosition   - position the tape to a specific block (aka seek)	*/
/*	TURead       - read one 512 byte logical block from the tape	*/
/*	TUWrite      - write one 512 byte logical block to the tape	*/
/*									*/
/*   Normally the TUInitialize and TUDiagnose routines would be called	*/
/* during the DOS device driver initialization, and then the TURead and	*/
/* TUWrite routines would be used to transfer data.  TUPosition can be	*/
/* used to implement either a tape rewind function or a tape tensioning	*/
/* operation.   The other functions (e.g. TUSend, TUReceive, TUEnd) are	*/
/* not usually needed.							*/
/*									*/
/*   If any errors occur, especially protocol errors, then the driver	*/
/* will usually want to re-synchronize communications with the TU58 by	*/
/* calling TUInitialize before attempting any other operations.		*/
/*									*/
/* Known bugs:								*/
/*	1) If an error occurs during a READ, the TU58 will respond with	*/
/*	an END packet rather than DATA.  We don't deal with that.	*/
/*	2) If an error occurs during a WRITE, the TU58 will send an END	*/
/*	packet rather than CONTINUE.  We don't deal with that, either.	*/
/*									*/
/*				Bob Armstrong [1-Aug-94]		*/


#include <stdio.h>		/* needed for NULL, etc			*/
#include <mem.h>		/* memset, memcopy, etc			*/
#include "standard.h"		/* all definitions for this project	*/
#include "serial.h"		/* serial port communications routines	*/
#include "tu58.h"		/* device protocol and data defintions	*/


/* new_command */
/*   This routine will initialize a command packet to all zeros and 	*/
/* then fill in the opcode byte.					*/
/*									*/
/* INPUTS:								*/
/*	command	- address of the command packet				*/
/*	opcode	- desired opcode for the new command			*/
/*									*/
/* OUTPUTS:	a clean command packet with the opcode filled in	*/
/*									*/
PRIVATE void new_command (CMDPKT *command, BYTE opcode)
{
#pragma intrinsic memset
  memset((char *) command, 0, CMDPKTLEN);
  command->opcode = opcode;
}


/* TUSend */
/*   This routine will send a complete packet to the TU58, including	*/
/* all the necessary protocol bits. It will add the packet type (called	*/
/* "flags" in the TU58 documentation), the message byte count, and the	*/
/* correct checksum word.  The latter will be computed as we transmit	*/
/* the packet!								*/
/*									*/
/* INPUTS:								*/
/*	type	- the packet type (flags) byte				*/
/*	buffer	- address of the buffer containing the data to transmit	*/
/*	length	- length, in bytes, of the data (must be .LE. DATASIZE)	*/
/*									*/
PUBLIC void TUSend (BYTE type, BYTE far *buffer, BYTE length)
{
  WORD checksum, data;  int i;
  COMPutW(type);  COMPutW(length);
  checksum = (length << 8)  |  type;
  for (i = 0;  i < length;  ++i) {
    if (ODD(i)) {
      data |= *buffer << 8;  ADC(checksum, data);
    } else
      data = *buffer;
    COMPutW(*buffer++);
  }
  if (ODD(length))  ADC(checksum, data);
  COMPutW(LOW(checksum));  COMPutW(HIGH(checksum));
}


/* TUReceive */
/*   This routine reads a complete packet from the TU58 and performs	*/
/* the necessary protocol checking.  It will verify that the packet has	*/
/* the expected type and length, and that the checksum is correct.  It	*/
/* uses only timed serial read functions to guarantee that we can never	*/
/* hang here if the TU58 should die.  If the data from the TU58 fails	*/
/* any of these tests, then an error code is returned.			*/
/*									*/
/*   The caller specifies the timeout for the first byte read.  This is	*/
/* normally a LONGWAIT for commands that require tape motion, and a	*/
/* SHORTWAIT for commands which do not.  All bytes after the first are	*/
/* always timed with a SHORTWAIT.					*/
/*									*/
/*   There's a special case to make this routine more flexible: if the	*/
/* type parameter is zero, then the caller has already read the packet	*/
/* type (flags) byte.  In this case TUReceive will expect the next byte	*/
/* to be the message byte count (the second byte of a normal packet).	*/
/*									*/
/* INPUTS:								*/
/*	type	- the expected packet type (zero if already read)	*/
/*	buffer	- address of the buffer to receive the data		*/
/*	length	- expected length of the data from the TU58		*/
/*	timeout	- maximum time to wait for an initial response		*/
/*									*/
/* RETURNS:								*/
/*	0 (E_SUCCESS)	- everything's fine				*/
/*	E_TIMEOUT	- no response from the TU58			*/
/*	E_BADTYPE	- TU58 responded with the wrong packet		*/
/*	E_BADLENGTH	- TU58 data length was not what we expected	*/
/*	E_CHECKSUM	- packet checksum doesn't agree			*/
/*									*/
PUBLIC int TUReceive (BYTE type, BYTE far *buffer, BYTE length, int timeout)
{
  BYTE ch;  WORD checksum, data;  int i;
  if (type != 0) {
    if (!COMGetW(&ch, timeout))  return E_TIMEOUT;
    if (ch != type)  return E_BADTYPE;
  }
  if (!COMGetW(&ch, SHORTWAIT))  return E_TIMEOUT;
  if (ch != length)  return E_BADLENGTH;
  checksum = (length << 8)  |  type;
  for (i = 0;  i < length;  ++i) {
    if (!COMGetW(&ch, SHORTWAIT))  return E_TIMEOUT;
    if (ODD(i)) {
      data |= ch << 8;  ADC(checksum, data);
    } else
      data = ch;
    *buffer++ = ch;
  }
  if (ODD(length))  ADC(checksum, data);
  if (!COMGetW(&ch, SHORTWAIT))  return E_TIMEOUT;
  if (ch != LOW(checksum))  return E_CHECKSUM;
  if (!COMGetW(&ch, SHORTWAIT))  return E_TIMEOUT;
  if (ch != HIGH(checksum))  return E_CHECKSUM;
  return E_SUCCESS;
}


/* TUInitialize */
/*   This routine initializes the TU58 controller and puts the protocol	*/
/* into a known state.  It works by causing a BREAK condition on the	*/
/* transmitted data, which in theory should always be able to get the	*/
/* TU58's attention no matter what it is doing.  After the break it	*/
/* transmits an INITIALIZE command and waits for the TU58 to respond.	*/
/* If there's no answer (e.g. the TU58 is turned off, disconnected or	*/
/* just plain dead) then FALSE is returned.				*/
/*									*/
/*   This routine expects that the caller has already initialized the	*/
/* serial port!								*/
/*									*/
/*   NOTE: a bug (at least _I_ think it's a bug!) in the TU58 requires	*/
/* us to always send _two_ INITIALIZE commands. The first one is always	*/
/* lost by the controller...						*/
/*									*/
/* RETURNS:	TRUE if the TU58 is alive and listening			*/
/*									*/
PUBLIC BOOLEAN TUInitialize (void)
{
  int count;  BYTE ch;
  for (count = 0;  count < 8;  ++count) {
    COMBreak();
    COMPutW(F_INITIALIZE);  COMPutW(F_INITIALIZE);
    if (COMGetW(&ch, SHORTWAIT)  &&  ch == F_CONTINUE)  return TRUE;
  }
  return FALSE;
}


/* TUEnd */
/*   This routine will wait for and read an acknowledgement (aka END)	*/
/* packet from the TU58.  The amount of time it waits depends on the	*/
/* operation we just started and is specified by the timeout parameter.	*/
/* If there are any problems receiving the END packet it will return an	*/
/* appropriate error code - otherwise it returns the final operation	*/
/* status reported by the TU58 in the END packet.			*/
/*									*/
/* INPUTS:								*/
/*	timeout	  - amount of time (in ticks) to wait for a response	*/
/*									*/
/* RETURNS:	status from reading or as reported by the TU58		*/
/*									*/
PUBLIC int TUEnd (int timeout)
{
  ENDPKT response;  int status;
  status = TUReceive(F_CONTROL, (BYTE far *) &response, ENDPKTLEN, timeout);
  if (status != E_SUCCESS)  return status;
  if (response.opcode != O_END)  return E_BADTYPE;
  return response.success;
}


/* TUDiagnose */
/*   This routine asks the TU58 to execute its built in self test,	*/
/* waits for the response, and then returns the status reported by the	*/
/* TU58.  Note that the self test is non-destructive, doesn't involve	*/
/* any tape motion, and requires about 1/2 second to complete.		*/
/*									*/
/* RETURNS:	status code reported by the TU58 (or internal error)	*/
/*									*/
PUBLIC int TUDiagnose (void)
{
  CMDPKT command;
  new_command(&command, O_DIAGNOSE);
  TUSend(F_CONTROL, (BYTE far *) &command, CMDPKTLEN);
  return TUEnd(SHORTWAIT);
}


/* TUPosition */
/*   This routine will position the tape at the start of the specified	*/
/* logical block.  It isn't needed for reading or writing data, because	*/
/* the TU58 controller is smart enough to position automatically.  It	*/
/* is used to rewind the tape, and to do an end-to-end tape tensioning	*/
/* pass.								*/
/*									*/
/* INPUTS:								*/
/*	unit	- selects unit 0 or 1					*/
/*	lbn	- logical block number					*/
/*									*/
/* RETURNS:	status code reported by the TU58 (or internal error)	*/
/*									*/
PUBLIC int TUPosition (int unit, int lbn)
{
  CMDPKT command;
  new_command(&command, O_POSITION);
  command.unit = unit;  command.lbn_h = HIGH(lbn);  command.lbn_l = LOW(lbn);
  TUSend(F_CONTROL, (BYTE far *) &command, CMDPKTLEN);
  return TUEnd(LONGWAIT);
}


/* TURead */
/*   This routine will read one logical block from the TU58 and pack	*/
/* the data into the caller's buffer.  All the necessary error checking	*/
/* is done and the data can be assumed good.  The TU58 requires us to	*/
/* send a READ command with the appropriate information, and then it	*/
/* will return four DATA packets, each containing 128 bytes.  After the	*/
/* last DATA packet is an END packet containing the final status of the	*/
/* operation.								*/
/*									*/
/*  IMPORTANT!  Blocks are always 512 bytes!  Never more, never less.	*/
/*									*/
/* INPUTS:								*/
/*	unit	- selects tape drive 0 (left) or 1 (right)		*/
/*	lbn	- logical block number to be read [0..511]		*/
/*	buffer	- address of 512 bytes to receive the data read		*/
/*									*/
/* RETURNS:	operation status as reported by the TU58		*/
/*									*/
PUBLIC int TURead (int unit, int lbn, BYTE far *buffer)
{
  CMDPKT command;  int i, status;  BYTE response;
  new_command(&command, O_READ);
  command.unit = unit;
  command.lbn_l = LOW(lbn);        command.lbn_h = HIGH(lbn);
  command.dbc_l = LOW(BLOCKSIZE);  command.dbc_h = HIGH(BLOCKSIZE);
  TUSend(F_CONTROL, (BYTE far *) &command, CMDPKTLEN);
  for (i = 0;  i < BLOCKSIZE/DATASIZE;  ++i) {
    if (!COMPeekW(&response, LONGWAIT))  return E_TIMEOUT;
    if (response == F_CONTROL)  return TUEnd(SHORTWAIT);
    status = TUReceive(F_DATA, buffer, DATASIZE, SHORTWAIT);
    if (status != E_SUCCESS)  return status;
    buffer += DATASIZE;
  }
  return TUEnd(SHORTWAIT);
}


/* TUWrite */
/*   This routine will write one logical block to the TU58.  To write	*/
/* data the TU58 wants us to first send a WRITE packet and then four	*/
/* DATA packets, each containing 128 bytes. In between the DATA packets	*/
/* we must wait for a CONTINUE message from the TU58.  This complex	*/
/* arrangement prevents overflowing the poor TU58 controller, which has	*/
/* only 128 bytes of buffer space!  After the last data block, the TU58	*/
/* will send us an END packet containing the final operation status.	*/
/*									*/
/*   The TU58 controller has the option to read back each 128 byte 	*/
/* block after it is written and compare it to the original data.  This	*/
/* read is performed with a reduced sensitivity on the read amplifier	*/
/* and will detect any bad spots on the tape.  This gives us the option	*/
/* of a "write with verification" operation.				*/
/*									*/
/*  IMPORTANT!  Blocks are always 512 bytes!  Never more, never less.	*/
/*									*/
/* INPUTS:								*/
/*	unit	- selects tape drive 0 (left) or 1 (right)		*/
/*	lbn	- logical block number to be read [0..511]		*/
/*	buffer	- address of 512 bytes containing the data to write	*/
/*	verify	- TRUE to ask the TU58 for a verification pass		*/
/*									*/
/* RETURNS:	operation status as reported by the TU58		*/
/*									*/
PUBLIC int TUWrite (int unit, int lbn, BYTE far *buffer, BOOLEAN verify)
{
  CMDPKT command;  int i, status;  BYTE response;
  new_command(&command, O_WRITE);
  command.unit = unit;
  command.lbn_l = LOW(lbn);        command.lbn_h = HIGH(lbn);
  command.dbc_l = LOW(BLOCKSIZE);  command.dbc_h = HIGH(BLOCKSIZE);
  if (verify)  command.modifier = M_CHECK;
  TUSend(F_CONTROL, (BYTE far *) &command, CMDPKTLEN);
  for (i = 0;  i < BLOCKSIZE/DATASIZE;  ++i) {
    if (!COMPeekW(&response, LONGWAIT))  return E_TIMEOUT;
    if (response == F_CONTROL)  break;
    if (response != F_CONTINUE)  return E_BADTYPE;
    COMGet(&response);
    TUSend(F_DATA, buffer, DATASIZE);
    buffer += DATASIZE;
  }
  /*   Normally the TU58 will return the end packet immediately, UNLESS	*/
  /* we are doing a write-with-verify function. Then the TU58 has to go	*/
  /* back and re-read what it just wrote.  That can take a while!	*/
  return TUEnd(verify ? LONGWAIT : SHORTWAIT);
}
